// CS370 - Fall 2016
// Lab20 - Mirror Robot

#ifdef OSX
	#include <GLUT/glut.h>
#else
	#include <GL/glew.h>
	#include <GL/glut.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <SOIL/SOIL.h>
#include "materials.h"

// Shader file utility functions
#include "shaderutils.h"

// Shader files
GLchar* texVertexFile = "texturevert.vs";
GLchar* texFragmentFile = "texturefrag.fs";
GLchar* defaultVertexFile = "basicvert.vs";
GLchar* defaultFragmentFile = "basicfrag.fs";

// Shader objects
GLuint textureProg;
GLuint defaultProg;
GLint texSampler;

// Node structure
struct treenode
{
	MaterialType material;
	GLfloat color[4];
	GLuint texture;
	void (*f)();
	GLfloat m[16];
	treenode *sibling;
	treenode *child;
	GLuint shaderProg;
};

// Texture constants
#define NO_TEXTURES 2
#define ENVIRONMENT 0
#define METAL 1
#define X 0
#define Y 1
#define Z 2

// Texture indices
GLuint tex_ids[NO_TEXTURES];

// Texture files
char texture_files[NO_TEXTURES][20] = {"blank.bmp","metal.jpg"};

// Define object dimensions
#define BASE_RADIUS 3.0f
#define BASE_HEIGHT 2.0f
#define LOWER_HEIGHT 4.0f
#define LOWER_WIDTH 2.0f
#define LOWER_DEPTH 1.0f
#define UPPER_HEIGHT 3.0f
#define UPPER_WIDTH 1.0f
#define UPPER_DEPTH 1.0f

// Scene nodes
treenode base;
treenode lower_arm;
treenode left_upper_arm;
treenode right_upper_arm;

// Rotation angles
GLfloat theta = 0.0f;
GLfloat dtheta = 1.0f;
GLfloat phi = 0.0f;
GLfloat dphi = 1.0f;
GLfloat left_psi = 0.0f;
GLfloat right_psi = 0.0f;
GLfloat dpsi = 1.0f;

// Quadric object
GLUquadricObj *basequad;

// Global camera vectors
GLfloat eye[3] = {1.0f,1.0f,1.0f};
GLfloat at[3] = {0.0f,0.0f,0.0f};
GLfloat up[3] = {0.0f,1.0f,0.0f};

// Global screen dimensions
GLfloat ww,hh;

void display();
void render_Scene();
void create_Mirror();
void render_Mirror();
void keyfunc(unsigned char key, int x, int y);
void reshape(int w, int h);
bool load_textures();
void build_scene_graph();
void traverse(treenode *node);
void update_base();
void update_lower_arm();
void update_left_upper_arm();
void update_right_upper_arm();
void draw_base();
void draw_lower_arm();
void draw_upper_arm();

int main(int argc, char *argv[])
{
	// Initialize GLUT
	glutInit(&argc,argv);

	// Initialize the window with double buffering and RGB colors
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	// Set the window size to image size
	glutInitWindowSize(512,512);

	// Create window
	glutCreateWindow("Mirror Robot");

#ifndef OSX
	// Initialize GLEW - MUST BE DONE AFTER CREATING GLUT WINDOW
	glewInit();
#endif

	// Define callbacks
	glutDisplayFunc(display);
	glutKeyboardFunc(keyfunc);
	glutReshapeFunc(reshape);

	// Set background color
	glClearColor(1.0f,1.0f,1.0f,1.0f);

	// Enable depth test
	glEnable(GL_DEPTH_TEST);

	// Create base quadric with texture coords
	basequad = gluNewQuadric();
	gluQuadricDrawStyle(basequad,GLU_FILL);
	gluQuadricTexture(basequad,GL_TRUE);

	// Load textures
	if (!load_textures())
	{
		exit(0);
	}

	// Load shader programs
	textureProg = load_shaders(texVertexFile,texFragmentFile);
	defaultProg = load_shaders(defaultVertexFile,defaultFragmentFile);

	// Associate and assign sampler parameter
	texSampler = glGetUniformLocation(textureProg,"texMap");

	// Build scene graph
	build_scene_graph();

	// Begin event loop
	glutMainLoop();

	return 0;
}

// Display callback
void display()
{
	// TODO: PASS 1 - Create mirror texture

	// PASS 2 - Render actual scene
	// Reset background
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Set projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// Adjust viewing volume (orthographic)
	GLfloat xratio = 1.0f;
	GLfloat yratio = 1.0f;
	// If taller than wide adjust y
	if (ww <= hh)
	{
		yratio = (GLfloat)hh / (GLfloat)ww;
	}
	// If wider than tall adjust x
	else if (hh <= ww)
	{
		xratio = (GLfloat)ww / (GLfloat)hh;
	}
	glOrtho(-10.0f*xratio,10.0f*xratio,-5.0f*yratio,10.0f*yratio,-10.0f,10.0f);

	// Set modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eye[X], eye[Y], eye[Z], at[X], at[Y], at[Z], up[X], up[Y], up[Z]);

	// Render scene
	render_Scene();

	// TODO: Render mirror

	// Flush buffer
	glFlush();

	// Swap buffers
	glutSwapBuffers();
}

// Scene render function
void render_Scene()
{
	// Traverse scene graph
	glPushMatrix();
		traverse(&base);
	glPopMatrix();
}

void create_Mirror()
{
	// PASS 1 - Render reflected scene
	// Reset background
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// TODO: Set projection matrix for flat "mirror" camera
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// TODO: Set modelview matrix positioning "mirror" camera
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// TODO: Render scene from mirror

	glFinish();

	// TODO: Copy scene to texture

}

// Mirror render function
void render_Mirror()
{
	glPushMatrix();
		glUseProgram(textureProg);
		glUniform1i(texSampler,0);
		// TODO: Draw mirror surface
		
		glUseProgram(defaultProg);
		glColor3f(0.0f,0.0f,0.0f);
		// Draw mirror frame
		glBegin(GL_LINE_LOOP);
			glVertex3f(-7.0f, -3.0f, 2.0f);
			glVertex3f(-7.0f, -3.0f, -6.0f);
			glVertex3f(-7.0f, 5.0f, -6.0f);
			glVertex3f(-7.0f, 5.0f, 2.0f);
		glEnd();

	glPopMatrix();
}

// Routine to load textures using SOIL
bool load_textures()
{
	// Load environment map texture (NO MIPMAPPING)
	tex_ids[ENVIRONMENT] = SOIL_load_OGL_texture(texture_files[0],SOIL_LOAD_AUTO,SOIL_CREATE_NEW_ID,SOIL_FLAG_INVERT_Y);

	// TODO: Set environment map properties if successfully loaded
	if (tex_ids[ENVIRONMENT] != 0)
	{
		// Set scaling filters (no mipmap)

		// Set wrapping modes (clamped)

	}
	// Otherwise texture failed to load
	else
	{
		return false;
	}

	// Load object textures normally
	for (int i=1; i < NO_TEXTURES; i++)
	{
		// Load images
		tex_ids[i] = SOIL_load_OGL_texture(texture_files[i],SOIL_LOAD_AUTO,SOIL_CREATE_NEW_ID,SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y);


		// Set texture properties if successfully loaded
		if (tex_ids[i] != 0)
		{
			// Set scaling filters
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);

			// Set wrapping modes
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		}
		// Otherwise texture failed to load
		else
		{
			return false;
		}
	}
	return true;
}

// Keyboard callback
void keyfunc(unsigned char key, int x, int y)
{
	// a d rotate base
	if (key == 'a')
	{
		theta += dtheta;
		if (theta > 360.0f)
		{
			theta -= 360.0f;
		}
		update_base();
	}
	else if (key == 'd')
	{
		theta -= dtheta;
		if (theta < 0.0f)
		{
			theta += 360.0f;
		}
		update_base();
	}

	// w s rotates lower arm
	if (key == 'w')
	{
		phi += dphi;
		if (phi > 90.0f)
		{
			phi = 90.0f;
		}
		update_lower_arm();
	}
	else if (key == 's')
	{
		phi -= dphi;
		if (phi < -90.0f)
		{
			phi = -90.0f;
		}
		update_lower_arm();
	}

	// m n rotates left upper arm
	if (key == 'm')
	{
		left_psi += dpsi;
		if (left_psi > 180.0f)
		{
			left_psi = 180.0f;
		}
		update_left_upper_arm();
	}
	else if (key == 'n')
	{
		left_psi -= dpsi;
		if (left_psi < -180.0f)
		{
			left_psi = -180.0f;
		}
		update_left_upper_arm();
	}

	// . , rotates right upper arm
	if (key == '.')
	{
		right_psi += dpsi;
		if (right_psi > 180.0f)
		{
			right_psi = 180.0f;
		}
		update_right_upper_arm();
	}
	else if (key == ',')
	{
		right_psi -= dpsi;
		if (right_psi < -180.0f)
		{
			right_psi = -180.0f;
		}
		update_right_upper_arm();
	}

	// <esc> quits
	if (key == 27)
	{
		exit(0);
	}

	// Redraw screen
	glutPostRedisplay();
}

// Routine to build scene graph
void build_scene_graph()
{
	// TODO: Set base node texture
	
	base.f = draw_base;
	base.sibling = NULL;
	base.child = &lower_arm;
	base.shaderProg = textureProg;
	update_base();

	// Create lower arm node
	lower_arm.color[0] = 1.0f;
	lower_arm.color[1] = 0.0f;
	lower_arm.color[2] = 0.0f;
	lower_arm.color[3] = 1.0f;
	lower_arm.f = draw_lower_arm;
	lower_arm.sibling = NULL;
	lower_arm.child = &left_upper_arm;
	lower_arm.shaderProg = defaultProg;
	update_lower_arm();

	// Create left upper arm node
	left_upper_arm.color[0] = 0.0f;
	left_upper_arm.color[1] = 0.0f;
	left_upper_arm.color[2] = 1.0f;
	left_upper_arm.color[3] = 1.0f;
	left_upper_arm.f = draw_upper_arm;
	left_upper_arm.sibling = &right_upper_arm;
	left_upper_arm.child = NULL;
	left_upper_arm.shaderProg = defaultProg;
	update_left_upper_arm();

	// Create right upper arm node
	right_upper_arm.color[0] = 0.0f;
	right_upper_arm.color[1] = 0.0f;
	right_upper_arm.color[2] = 1.0f;
	right_upper_arm.color[3] = 1.0f;
	right_upper_arm.f = draw_upper_arm;
	right_upper_arm.sibling = NULL;
	right_upper_arm.child = NULL;
	right_upper_arm.shaderProg = defaultProg;
	update_right_upper_arm();
}

// Routine to traverse scene graph
void traverse(treenode *node)
{
	// Stop when at bottom of branch
	if (node == NULL)
	{
		return;
	}

	// Apply local transformation and render
	glPushMatrix();
	glMultMatrixf(node->m);
	glUseProgram(node->shaderProg);
	node->f();

	// Recurse vertically if possible (depth-first)
	if (node->child != NULL)
	{
		traverse(node->child);
	}

	// Remove local transformation and recurse horizontal
	glPopMatrix();
	if (node->sibling != NULL)
	{
		traverse(node->sibling);
	}
}

// Routine to update base transformation
void update_base()
{
	glPushMatrix();
		glLoadIdentity();
		// Rotate about y axis
		glRotatef(theta,0.0f,1.0f,0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX,base.m);
	glPopMatrix();
}

// Routine to update lower arm transformation
void update_lower_arm()
{
	glPushMatrix();
		glLoadIdentity();
		// Move on top of base
		glTranslatef(0.0f,BASE_HEIGHT,0.0f);
		// Rotate about x axis
		glRotatef(phi,1.0f,0.0f,0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX,lower_arm.m);
	glPopMatrix();
}

// Routine to update left upper arm transformation
void update_left_upper_arm()
{
	glPushMatrix();
		glLoadIdentity();
		// Move to left side of lower arm
		glTranslatef((LOWER_WIDTH+UPPER_WIDTH)/2.0f,LOWER_HEIGHT,0.0f);
		// Rotate about x axis
		glRotatef(left_psi,1.0f,0.0f,0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX,left_upper_arm.m);
	glPopMatrix();
}

// Routine to update right upper arm transformation
void update_right_upper_arm()
{
	glPushMatrix();
		glLoadIdentity();
		// Move to right side of lower arm
		glTranslatef(-(LOWER_WIDTH+UPPER_WIDTH)/2.0f,LOWER_HEIGHT,0.0f);
		// Rotate about x axis
		glRotatef(right_psi,1.0f,0.0f,0.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX,right_upper_arm.m);
	glPopMatrix();
}

// Routine to draw base
void draw_base()
{
	glPushMatrix();
		// TODO: Set texture

		// Rotate onto x-y plane
		glRotatef(-90.0f,1.0f,0.0f,0.0f);
		// Draw quadric
		gluCylinder(basequad,BASE_RADIUS,BASE_RADIUS,BASE_HEIGHT,20,20);
	glPopMatrix();
}

// Routine to draw lower arm
void draw_lower_arm()
{
	glPushMatrix();
		glColor3fv(lower_arm.color);
		// Translate up to x-y plane
		glTranslatef(0.0f,LOWER_HEIGHT/2.0f,0.0f);
		// Scale
		glScalef(LOWER_WIDTH,LOWER_HEIGHT,LOWER_DEPTH);
		glutSolidCube(1.0f);
	glPopMatrix();
}

// Routine to draw lower arm
void draw_upper_arm()
{
	glPushMatrix();
		glColor3fv(left_upper_arm.color);
		// Translate up to x-y plane
		glTranslatef(0.0f,UPPER_HEIGHT/2.0f,0.0f);
		// Scale
		glScalef(UPPER_WIDTH,UPPER_HEIGHT,UPPER_DEPTH);
		glutSolidCube(1.0f);
	glPopMatrix();
}

// Reshape callback
void reshape(int w, int h)
{
	// Set new screen extents
	glViewport(0, 0, w, h);

	// Store new extents
	ww = w;
	hh = h;
}
